unit DocManager;

{$mode objfpc}{$H+}

interface

uses
  Classes, Controls, StdCtrls, SynEdit, LCLType, ExtCtrls, Forms,
  Xscontrols, Graphics, Dialogs, SysUtils, LConvEncoding, Buttons,
  SynGutter, SynGutterCodeFolding, SynGutterLineNumber, LCLIntf, ComCtrls,
  SynHighlighterObr, SynEditKeyCmds, SynEditMarkupHighAll, XsEditors,
  SynEditMouseCmds, SynEditHighlighter, SynEditTypes, Opts,
  DOM, XMLWrite, XMLRead;//, Parser;

const
  sep = DirectorySeparator;

type

  TSpecPage = (spHexview,spCompare);

  { TXSSynEdit }

  TXSSynEdit = class(TSynEdit)
  public
    encode    : string;
    ErrorLine : integer;
    DebugLine : integer;
    SynMarkup : TSynEditMarkupHighlightAllCaret;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published

  end;

  { TXSDockManager }

  TXSDocManager = class
  private
    ftab        : TXSScrollBox;
    fdown       : boolean;
    feditor     : TXSSynEdit;
    fdefname    : string;
    fdefext     : string;
    fprjname    : string;
    fprjext     : string;
    ffilter     : string;
    wordatxy    : string;
    searchxy    : TPoint;
    fopendialog : TOpenDialog;
    fsavedialog : TSaveDialog;
    ftreeview   : TXSTreeView;
    fpagecontrol: TXSpagecontrol;
    fsearchopt  : TSynSearchOptions;
    foldtext    : string;
    procedure   PageChange(Sender: TObject);
    function    GetEditorByNum(num,idx: integer): TXSSynEdit;
    procedure   EditorOptions(ed: TXSSynEdit; tab: TXSScrollBox);
    procedure   InvalidateParent(p: TControl);
    function    GetTabByNum(num, idx: integer): TXSScrollBox;
    procedure   EditorChange(Sender: TObject);
    procedure   ScrollChange(Sender: TObject);
    procedure   SwapCommand(idx1,idx2: TSynEditorMouseCommand);
    procedure   EditorKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure   EditorKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure   EditorPaint(Sender: TObject; ACanvas: TCanvas);
    procedure   MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   EditorMouseLink(Sender: TObject; X, Y: Integer; var AllowMouseLink: Boolean);
    procedure   EditorClickLink(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
    procedure   SpecialLineColors(Sender: TObject; Line: integer; var Special: boolean; var FG, BG: TColor);
    procedure   OpenPageByName(name: string);
    function    FindPageByName(name: string): integer;
    procedure   SaveText(idx: integer; path, from, encode: string);
    procedure   SearchReplace;
    procedure   PageClose(Sender: TObject; idx: integer);

    procedure   HexOpenClick(Sender: TObject);
    function    GetHexByTag(tag: integer): TXSHexView;
    procedure   GetHexEditByTag(parent: twincontrol; tag: integer; var ed: TEdit);
    procedure   HexReload(Sender: TObject);
    procedure   CompareClick(Sender: TObject);
    procedure   PanelResize(Sender: TObject);
    procedure   TabResize(Sender: TObject);
    procedure   ComparePanelResize(Sender: TObject);
  public
    constructor Create(tree: TXSTreeView; page: TXSpagecontrol);
    destructor  Destroy; override;
    procedure   UpdateText;
    function    TabIsSpecial: boolean;
    function    TabPath: string;
    function    AllModified: boolean;
    function    PageCount: integer;

    procedure   NewSpecPage(mode: TSpecPage; name,tabname: string);
    procedure   NewProject;
    procedure   SaveXmlProject(node: TTreeNode; name: string);
    procedure   OpenXmlProject(node: TTreeNode; name: string);
    procedure   OpenProject;
    function    SaveProjectNode(node: TTreeNode): Boolean;
    procedure   SaveProjectAs;
    function    SaveProjectByNode(node: TTreeNode): boolean;
    procedure   SaveProject;
    function    CloseProjectNode(node: TTreeNode): Boolean;
    procedure   CloseProject;
    procedure   SaveProjectFromMenu;
    procedure   SaveProjectAsFromMenu;
    procedure   SetAsMainFromMenu;
    function    PrjIsModified(flag: boolean): boolean;
    procedure   AddFilesInPrj;
    procedure   RemoveFileInPrj;
    function    SaveCloseQueryProjects: Boolean;
    procedure   AddDebugTab(name: string);
    procedure   TreeDblClick(Sender: TObject);
    procedure   TreeOnClick(Sender: TObject);
    function    AddModInPrjByNode(node: TTreeNode; path: string): TTreeNode;
    function    AddModInPrj(path: string): TTreeNode;
    procedure   SelectNode(path: string);
    procedure   PopupClosePage;
    procedure   PopupAddInPrj;
    procedure   PopupSetAsMain;

    function    PageTemplate(name: string): string;
    procedure   NewPage;
    procedure   OpenPage;
    procedure   ClosePage;
    function    SaveFileAsByIdx(idx: integer): Boolean;
    procedure   SaveFileAs;
    function    SaveFileByIdx(idx: integer): Boolean;
    procedure   SaveFile;
    procedure   SaveAll;
    function    CloseFileByIdx(idx: integer): Boolean;
    procedure   CloseFile;
    function    CloseAll: Boolean;
    function    SaveCloseQuery: Boolean;

    procedure   Cut;
    procedure   Copy;
    procedure   Paste;
    procedure   Delete;
    procedure   Undo;
    procedure   Redo;
    procedure   Find;
    procedure   FindNext;
    procedure   FindPrev;
    procedure   Unindent;
    procedure   Indent;
    procedure   Build;
    function    Run: boolean;
    procedure   Pause;
    procedure   Stop;
    procedure   SelectAll;
    procedure   DeleteSelection;
    procedure   Upper;
    procedure   Lower;
    procedure   Swap;
    procedure   TitleUp;
    procedure   TitleLo;
    procedure   Browse;

  published
    property    DefModName : string   read fdefname write fdefname;
    property    DefModExt  : string   read fdefext write fdefext;
    property    DefPrjName : string   read fprjname write fprjname;
    property    DefPrjExt  : string   read fprjext write fprjext;
    property    FileFilter : string   read ffilter write ffilter;
    property    Editor     : TXSSynEdit read feditor write feditor;
  end;

implementation

uses
  Main;

{ TXSSynEdit }

constructor TXSSynEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ErrorLine := -1;
  DebugLine := -1;
end;

destructor TXSSynEdit.Destroy;
begin

  inherited Destroy;
end;

{ TXSDocManager }

constructor TXSDocManager.Create(tree: TXSTreeView; page: TXSpagecontrol);
begin
  inherited Create;
  fdefname := 'Untitled';
  fdefext  := '.txt';
  fprjname := 'Project';
  fprjext  := '.prj';
  ffilter  := 'Text files (*.txt;*.ini;*.inc)|*.txt;*.ini;*.inc'
             +'|All files (*.*)|*.*';

  fopendialog           := TOpenDialog.Create(nil);
  fopendialog.Options   := fopendialog.Options + [ofAllowMultiSelect];
  fsavedialog           := TSaveDialog.Create(nil);

  ftreeview                := tree;
  ftreeview.OnMouseDown    := @MouseDown;
  ftreeview.OnDblClick     := @TreeDblClick;
  ftreeview.OnClick        := @TreeOnClick;

  fpagecontrol             := page;
  fpagecontrol.OnChange    := @PageChange;
  fpagecontrol.OnMouseDown := @MouseDown;
  fpagecontrol.OnPageClose := @PageClose;

  fpagecontrol.AddText('1:1',70);
  fpagecontrol.AddText('UTF8',60);
  fpagecontrol.AddText('',55);
  fpagecontrol.AddText('',250);

  fsearchopt := [ssoFindContinue];
  feditor := nil;

end;

destructor TXSDocManager.Destroy;
begin
  FreeAndNil(fopendialog);
  FreeAndNil(fsavedialog);
  inherited Destroy;
end;

procedure TXSDocManager.UpdateText;
begin
  if fpagecontrol.pagelist.count=0 then exit;
  if feditor=NIL then begin
    fpagecontrol.Text[0] := '';
    fpagecontrol.Text[1] := '';
    fpagecontrol.Text[2] := '';
    fpagecontrol.Text[3] := '';
  end else begin
    fpagecontrol.Text[0] := IntToStr(feditor.CaretY)+':'+IntToStr(feditor.CaretX);
    fpagecontrol.Text[1] := UpperCase(feditor.encode);
    fpagecontrol.Text[2] := '';
    if feditor.Modified then fpagecontrol.Text[2] := 'Modified';
    fpagecontrol.Text[3] := GetTabByNum(fpagecontrol.activepage,0).Path;
  end;
end;

procedure TXSDocManager.InvalidateParent(p: TControl);
begin
  while p.Parent<>NIL do p := p.Parent;
  p.Invalidate;
end;

function TXSDocManager.GetTabByNum(num, idx: integer): TXSScrollBox;
begin
  result := NIL;
  if (num<fpagecontrol.pagelist.count) and (idx<TStringList(fpagecontrol.pagelist.objects[num]).count) then
    result := TXSScrollBox(TStringList(fpagecontrol.pagelist.objects[num]).objects[idx]);
end;

procedure TXSDocManager.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Sender IS TXSScrollbox   then TXSScrollbox(Sender).SetFocus;
  if Sender IS TXSpagecontrol then TXSpagecontrol(Sender).SetFocus;
  if (Sender IS TXSTreeView) then Form1.prjpanel.SetFocus;
  if Sender IS TXSSynEdit     then begin
    TXSSynEdit(Sender).SetFocus;
    TXSSynEdit(Sender).ErrorLine := 0;
    TXSSynEdit(Sender).SynMarkup.Enabled := true;
    feditor.SelectedColor.Background := Options.SynEdit1.SelectedColor.Background;
  end;
  UpdateText;
  InvalidateParent(fpagecontrol);
end;

procedure TXSDocManager.PageChange(Sender: TObject);
var
  i  : integer;
begin
  ftab := nil;
  if (fpagecontrol.pagelist.count > 0) then begin
    ftab := GetTabByNum(fpagecontrol.ActivePage,fpagecontrol.ActiveTab);
    if (ftab <> nil) then begin
      SelectNode(GetTabByNum(fpagecontrol.ActivePage,0).path);
      feditor := nil;
      i:=0;
      while i<=ftab.ControlCount-1 do begin
        if ftab.Controls[i] is TXSSynEdit then begin
          feditor := TXSSynEdit(ftab.Controls[i]);
          if not feditor.visible then feditor.visible := true;
          feditor.SetFocus;
          break;
        end;
        inc(i);
      end;
    end;
  end else begin
    ftab    := nil;
    feditor := nil;
    SelectNode('');
  end;
  if ftab <> nil then fpagecontrol.VisibleText := not ftab.Special and (fpagecontrol.ActiveTab = 0);
  UpdateText;
  InvalidateParent(fpagecontrol);
end;

procedure TXSDocManager.PageClose(Sender: TObject; idx: integer);
begin
  CloseFileByIdx(idx);
  PageChange(self);
end;

procedure TXSDocManager.SwapCommand(idx1,idx2: TSynEditorMouseCommand);
var
  i: integer;
begin
  if ftab       = nil then exit;
  if feditor    = nil then exit;
  for i:= 0 to feditor.MouseActions.Count - 1  do
    if feditor.MouseActions.Items[i].Command = idx1 then
      feditor.MouseActions.Items[i].Command := idx2;
end;

function TXSDocManager.TabIsSpecial: boolean;
begin
  result := false;
  if ftab       = nil then exit;
  result := ftab.Special;
end;

function TXSDocManager.TabPath: string;
begin
  result := '';
  if fpagecontrol.activepage>=0 then
    result := ExtractFilePath(GetTabByNum(fpagecontrol.activepage,0).Path);
end;

function TXSDocManager.PageCount: integer;
begin
  result := fpagecontrol.pagelist.count;
end;

function TXSDocManager.AllModified: boolean;
var
  tab  : TXSScrollBox;
  i,j  : integer;
begin
  result := false;
  i := 0;
  while (i<=fpagecontrol.pagelist.count-1) and not result do begin
    j := 0;
    while (j<=TStringList(fpagecontrol.pagelist.objects[i]).count-1) and not result do begin
      tab := GetTabByNum(i,j);
      if (tab <> NIL) then result := tab.Modified;
      inc(j);
    end;
    inc(i);
  end;
end;

procedure TXSDocManager.SearchReplace;
var
  Res: integer;
begin
  if ftab       = nil then exit;
  if feditor    = nil then exit;
  Res:=feditor.SearchReplace(Form1.Edit1.Text,Form1.Edit1.Text,fsearchopt);
  if Res>0 then begin
    feditor.SelectedColor.Background := Options.SynEdit1.SelectedColor.Background;
    feditor.TopLine  := feditor.CaretY-(feditor.LinesInWindow div 2)
  end else begin
      // 'Search word "'+Form1.Edit1.Text+'" not found!'#13#10+
      //'Please enter a different search term!';
    if fsearchopt*[ssoBackwards]=[] then
      Form1.RadioButton1.Checked:=true
    else
      Form1.RadioButton2.Checked:=true;
    feditor.SelectedColor.Background := clRed;
  end;
end;

function TXSDocManager.GetEditorByNum(num,idx: integer): TXSSynEdit;
var
  tab: TXSScrollbox;
  i: integer;
  f: boolean;
begin
  result := NIL;
  tab := GetTabByNum(num,0);
  if tab.Special then exit;
  i:=0; f := false;
  while i<=tab.ControlCount-1 do begin
    f :=  tab.Controls[i] is TXSSynEdit;
    if f then begin
      result := TXSSynEdit(tab.Controls[i]);
      break;
    end;
    inc(i);
  end;
end;

procedure TXSDocManager.EditorChange(Sender: TObject);
begin
  if ftab       = nil then exit;
  if feditor    = nil then exit;
  ftab.Modified := feditor.Modified;
  UpdateText;
  fpagecontrol.Invalidate;
end;

procedure TXSDocManager.ScrollChange(Sender: TObject);
begin
  if feditor   = nil then exit;
  feditor.TopLine  := ftab.VScroll.Position+1;
  feditor.LeftChar := ftab.HScroll.Position;
end;

procedure TXSDocManager.EditorPaint(Sender: TObject; ACanvas: TCanvas);
begin
  if ftab      = nil then exit;
  if feditor   = nil then exit;
  ftab.VScroll.Max      := feditor.Lines.Count-1;
  ftab.VScroll.Position := feditor.TopLine-1;
  ftab.VScroll.PageSize := feditor.Height div feditor.LineHeight-1;
  ftab.HScroll.Max      := feditor.MaxLeftChar;
  ftab.HScroll.Position := feditor.LeftChar;
  ftab.HScroll.PageSize := feditor.CharsInWindow;
end;

procedure TXSDocManager.EditorClickLink(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
var
  i,w,xx,xs,ys: integer;
  s           : string;
begin
  if ftab      = nil then exit;
  if feditor   = nil then exit;
  SwapCommand(emcMouseLink,emcNone);
  feditor.MouseOptions := [];
  ys := searchxy.y-2;
  xs := 0;
  while (ys>0) and (xs=0) do begin
    xs:=Pos(wordatxy,feditor.Lines[ys]);
    dec(ys);
  end;
  if xs>0 then begin
    w := feditor.TabWidth;
    s := feditor.Lines[ys+1];
    i := 1;
    xx:= 1;
    while (i<Length(s)) and (i<=xs-1) do begin
      if s[i]=#9 then inc(xx,w) else inc(xx);
      inc(i);
    end;
    feditor.CaretXY:=Point(xx,ys+2);
    feditor.TopLine:=ys+2-(feditor.LinesInWindow div 2);
    UpdateText;
    fpagecontrol.Invalidate;
  end;
end;

procedure TXSDocManager.EditorMouseLink(Sender: TObject; X, Y: Integer;var AllowMouseLink: Boolean);
var
  Attri: TSynHighlighterAttributes;
  toktype,start: integer;
begin
  if ftab      = nil then exit;
  if feditor   = nil then exit;
  AllowMouseLink := true;
  searchxy := Point(X,Y);
  feditor.GetHighlighterAttriAtRowColEx(Point(X,Y),wordatxy,toktype,start,Attri);
end;

procedure TXSDocManager.EditorKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  ed: TXSSynEdit;
begin
  if ftab      = nil then exit;
  if feditor   = nil then exit;
  fdown := true;
  feditor.SelectedColor.Background := Options.SynEdit1.SelectedColor.Background;
  ed := TXSSynEdit(Sender);
  if ssCtrl IN Shift then begin
    SwapCommand(emcNone,emcMouseLink);
    ed.MouseOptions := [emUseMouseActions,emShowCtrlMouseLinks];
  end;
  if Key=VK_F3 then begin
    if not Form1.findpanel.Visible then Form1.findpanel.Visible:=true;
    foldtext := Form1.Edit1.Text;
    if not SameText(Form1.Edit1.Text,ed.SelText) and (ed.SelText<>'') then
      Form1.Edit1.Text:=ed.SelText;
    if Form1.RadioButton2.Checked then
      FindNext
    else
      FindPrev;
  end;
  if Key=VK_F8 then begin
    if (feditor.DebugLine=-1) or (feditor.DebugLine>=feditor.Lines.Count) then feditor.DebugLine:=0;
    INC(feditor.DebugLine);
    feditor.SetFocus;
    feditor.Invalidate;
  end;
  ed.ErrorLine := 0;
  ed.SynMarkup.Enabled := true;
  UpdateText;
  fpagecontrol.Invalidate;
end;

procedure TXSDocManager.EditorKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if ftab      = nil then exit;
  if feditor   = nil then exit;
  if key=17 then begin
    SwapCommand(emcMouseLink,emcNone);
    feditor.MouseOptions := [];
  end;
  fdown := false;
end;

PROCEDURE TXSDocManager.SpecialLineColors(Sender: TObject; Line: integer; VAR Special: boolean; VAR FG, BG: TColor);
begin
  if ftab      = nil then exit;
  if feditor   = nil then exit;
  if TXSSynEdit(Sender).ErrorLine=Line then begin
    Special := TRUE;
    FG      := clWhite;
    BG      := clRed;
  end else if TXSSynEdit(Sender).DebugLine=Line then begin
    Special := TRUE;
    FG      := clWhite;
    BG      := $CC9900;
  end;
end;

procedure TXSDocManager.EditorOptions(ed: TXSSynEdit; tab: TXSScrollBox);
begin
  ed.Highlighter := Options.Highlighter;
  ed.PopupMenu   := Form1.PopupMenu1;
  ed.Options     := ed.Options + [eoDragDropEditing];

  //ed.Font.Name   := 'PT Mono';
  //ed.Font.Size   := 10;

  ed.Font.Quality:= fqNonAntialiased;
  ed.Width       := Form1.Width;
  ed.Height      := Form1.Height;
  ed.Parent      := tab;
  ed.Align       := alClient;
  ed.BorderStyle := bsNone;
  ed.ScrollBars  := ssNone;

  ed.BracketMatchColor.Foreground  := Options.SynEdit1.BracketMatchColor.Foreground;
  ed.LineHighlightColor.Background := Options.SynEdit1.LineHighlightColor.Background;
  ed.SelectedColor.Background      := Options.SynEdit1.SelectedColor.Background;
  ed.Gutter.Color                  := Options.SynEdit1.Gutter.Color;

  TSynGutterLineNumber(ed.Gutter.Parts.Part[1]).ShowOnlyLineNumbersMultiplesOf:=10;
  TSynGutterLineNumber(ed.Gutter.Parts.Part[1]).MarkupInfo.Foreground:=TSynGutterLineNumber(Options.SynEdit1.Gutter.Parts.Part[1]).MarkupInfo.Foreground;
  TSynGutterLineNumber(ed.Gutter.Parts.Part[1]).MarkupInfo.Background:=TSynGutterLineNumber(Options.SynEdit1.Gutter.Parts.Part[1]).MarkupInfo.Background;
  TSynGutterSeparator(ed.Gutter.Parts.Part[3]).MarkupInfo.Foreground:=ed.Color;  //Разделитель
  TSynGutterCodeFolding(ed.Gutter.Parts.Part[4]).MarkupInfo.Foreground:=TSynGutterCodeFolding(Options.SynEdit1.Gutter.Parts.Part[4]).MarkupInfo.Foreground;
  TSynGutterCodeFolding(ed.Gutter.Parts.Part[4]).MarkupInfo.Background:=TSynGutterCodeFolding(Options.SynEdit1.Gutter.Parts.Part[4]).MarkupInfo.Background;
  ed.FoldedCodeColor.FrameColor:=Options.SynEdit1.FoldedCodeColor.FrameColor;

  ed.OnMouseLink := @EditorMouseLink;
  ed.OnClickLink := @EditorClickLink;
  ed.OnMouseDown := @MouseDown;
  ed.OnPaint     := @EditorPaint;
  ed.OnChange    := @EditorChange;
  ed.OnKeyDown   := @EditorKeyDown;
  ed.OnKeyUp     := @EditorKeyUp;
  ed.OnSpecialLineColors:=@SpecialLineColors;
  ed.ErrorLine   := 0;

  ed.SynMarkup := TSynEditMarkupHighlightAllCaret(ed.MarkupByClass[TSynEditMarkupHighlightAllCaret]);
  ed.SynMarkup.MarkupInfo.FrameColor := $E5E5E5;
  ed.SynMarkup.MarkupInfo.Background := $E0E0E0;
  ed.SynMarkup.WaitTime := 100; // millisec
  ed.SynMarkup.Trim := True;     // no spaces, if using selection
  ed.SynMarkup.FullWord := True; // only full words If "Foo" is under caret, do not mark it in "FooBar"
  ed.SynMarkup.IgnoreKeywords := False;

  tab.HScroll.OnChange := @ScrollChange;
  tab.VScroll.OnChange := @ScrollChange;

  ed.MouseLinkColor.Style:=[fsUnderline];
  ed.MouseActions.Add;
  ed.MouseActions.Items[0].Shift        := [ssCtrl];
  ed.MouseActions.Items[0].ShiftMask    := [ssShift, ssAlt, ssCtrl];
end;

procedure TXSDocManager.HexReload(Sender: TObject);
var
  ed1,ed2: TEdit;
  hex1,hex2: TXSHexView;
begin
  ed1 := nil;
  GetHexEditByTag(ftab,0,ed1);
  ed2 := nil;
  GetHexEditByTag(ftab,1,ed2);
  hex1 := GetHexByTag(0);
  hex2 := GetHexByTag(1);
  case TButton(Sender).tag of
    0: if (hex1<>nil) and (ed1<>nil) and (ed1.text<>'') and FileExists(ed1.text) then hex1.LoadFromFile(ed1.text);
    1: if (hex2<>nil) and (ed2<>nil) and (ed2.text<>'') and FileExists(ed2.text) then hex2.LoadFromFile(ed2.text);
  end;
end;

function TXSDocManager.GetHexByTag(tag: integer): TXSHexView;
var
  i: integer;
  f: boolean;
begin
  result := NIL;
  i:=0; f := false;
  while i<=ftab.ControlCount-1 do begin
    f :=  (ftab.Controls[i] is TXSHexView) and (tag = TXSHexView(ftab.Controls[i]).tag);
    if f then begin
      result := TXSHexView(ftab.Controls[i]);
      break;
    end;
    inc(i);
  end;
end;

procedure TXSDocManager.GetHexEditByTag(parent: twincontrol; tag: integer; var ed: TEdit);
var
  i,j: integer;
  f: boolean;
  p: TPanel;
begin
  i:=0; f := false;
  while i<=parent.ControlCount-1 do begin
    if parent.controls[i].ComponentCount>0 then GetHexEditByTag(twincontrol(parent.controls[i]),tag,ed);
    f :=  (parent.Controls[i] is TEdit) and (tag = TEdit(parent.Controls[i]).tag);
    if f then begin
      ed := TEdit(parent.Controls[i]);
      break;
    end;
    inc(i);
  end;
end;

procedure TXSDocManager.HexOpenClick(Sender: TObject);
var
  hex1,hex2: TXSHexView;
  ed1,ed2  : TEdit;
  fn1,fn2: string;
begin
  if ftab       = nil then exit;
  fopendialog.Filter := ffilter;
  fopendialog.FilterIndex := 4;
  if fopendialog.Execute then begin

    if fopendialog.Files.Count>1 then begin
      fn1 := fopendialog.Files[0];
      fn2 := fopendialog.Files[1];
    end else begin
      case TButton(Sender).tag of
        0: begin fn2 := fopendialog.FileName; fn1 := ''; end;
        1: begin fn1 := fopendialog.FileName; fn2 := ''; end;
      end;
    end;

    hex1 := GetHexByTag(0);
    if (hex1<>nil) and (fn2<>'') then begin
      hex1.LoadFromFile(fn2);
      ed1 := nil;
      GetHexEditByTag(ftab,0,ed1);
      if ed1<>nil then ed1.text := fn2;
    end;
    hex2 := GetHexByTag(1);
    if (hex2<>nil) and (fn1<>'') then begin
      hex2.LoadFromFile(fn1);
      ed2 := nil;
      GetHexEditByTag(ftab,1,ed2);
      if ed2<>nil then ed2.text := fn1;
    end;

  end;
end;

procedure TXSDocManager.CompareClick(Sender: TObject);
var
  i,len: integer;
  hex1,hex2: TXSHexView;
begin
  hex1 := GetHexByTag(0);
  hex2 := GetHexByTag(1);
  len := hex1.Len;
  if (len<>hex2.Len) and (len > hex2.Len) then len := hex2.Len;
  for i:=0 to len-1 do begin
    hex1.ChangedArray[i] := hex1.Data[i] <> hex2.Data[i];
    hex2.ChangedArray[i] := hex1.ChangedArray[i];
  end;
  hex1.Invalidate;
  hex2.Invalidate;
end;

procedure TXSDocManager.PanelResize(Sender: TObject);
var
  i: integer;
begin
  for i:=0 to TPanel(Sender).ControlCount-1 do begin
    if  (TPanel(Sender).Controls[i] is TPanel)
    and (TPanel(TPanel(Sender).Controls[i]).align<>alClient) then
      TPanel(Sender).Controls[i].width := TPanel(Sender).width div 2;
  end;
end;

procedure TXSDocManager.TabResize(Sender: TObject);
var
  i: integer;
begin
  for i:=0 to TXSScrollbox(Sender).ControlCount-1 do begin
    if  (TXSScrollbox(Sender).Controls[i] is TXSHexView)
    and (TXSHexView(TXSScrollbox(Sender).Controls[i]).align<>alClient) then
      TXSScrollbox(Sender).Controls[i].width := TXSScrollbox(Sender).width div 2;
  end;
end;

procedure TXSDocManager.ComparePanelResize(Sender: TObject);
var
  i: integer;
begin
  for i:=0 to TPanel(Sender).ControlCount-1 do begin
    if  (TPanel(Sender).Controls[i] is TButton) then
      TPanel(Sender).Controls[i].left := (TPanel(Sender).width - TPanel(Sender).Controls[i].width) div 2;
  end;
end;

procedure TXSDocManager.NewSpecPage(mode: TSpecPage; name,tabname: string);
var
  tab   : TXSScrollbox;
  panel : TPanel;
  button: TButton;
  ed1,ed2: TEdit;
  hex1,hex2: TXSHexView;

  procedure CreatePanel(parent: TWinControl;var edit: TEdit; align: TAlign; w,t: integer);
  var
    pan: TPanel;
    btn: TButton;
    rep: TSpeedButton;
    shp: TShape;
  begin
    pan := TPanel.Create(parent);
    pan.parent := parent;
    pan.align := align;
    pan.width := w;
    pan.Height:= 35;
    pan.color := cl3DLight;
    pan.BevelOuter:=bvNone;

    rep:= TSpeedButton.Create(pan);
    rep.parent := pan;
    rep.Anchors:= [akRight];
    rep.caption := '';
    rep.width := 25;
    rep.height := 25;
    rep.left := pan.width - rep.width-5;
    rep.top := 5;
    if Form1.ImageList1<>NIL then Form1.ImageList1.GetBitmap(42,rep.Glyph) ;
    rep.onclick := @HexReload;
    rep.tag := t;
    rep.showhint := true;
    rep.hint := 'Refresh...';

    btn:= TButton.Create(pan);
    btn.parent := pan;
    btn.Anchors:= [akRight];
    btn.caption := 'Open';
    btn.width := 45;
    btn.height := 25;
    btn.left := pan.width - rep.width-btn.width-5;
    btn.top := 5;
    btn.onclick := @HexOpenClick;
    btn.tag := t;

    shp:= TShape.Create(pan);
    shp.Parent:=pan;
    shp.Shape := stRoundRect;
    shp.Brush.color := clWhite;
    shp.Pen.color := clSilver;
    shp.left := 5;
    shp.top := 5;
    shp.height := 25;
    shp.width:= btn.left - 5;
    shp.Anchors:= [akLeft,akRight];

    edit:= TEdit.Create(pan);
    edit.parent := pan;
    edit.BorderStyle:=bsNone;
    edit.top  := shp.top+5;
    edit.left := shp.left+2;
    edit.height := shp.height-8;
    edit.width := shp.width-4;
    edit.Anchors:= [akLeft,akRight];
  end;

begin
  tab := fpagecontrol.AddPage(name,tabname,'',false);
  tab.Special := true;
  tab.VScroll.visible := false;
  tab.HScroll.visible := false;
  tab.OnMouseDown := @MouseDown;
  tab.OnResize:=@TabResize;

  panel := TPanel.Create(tab);
  panel.parent := tab;
  panel.align := alTop;
  panel.Height:= 35;
  panel.color := cl3DLight;
  panel.BevelOuter:=bvNone;
  panel.BorderStyle:=bsNone;
  panel.OnResize:=@PanelResize;

  if mode=spHexview then begin
    CreatePanel(panel,ed1,alClient,panel.width,0);
    ed1.tag := 0;
    hex1 := TXSHexView.Create(tab);
    hex1.tag:= 0;
    hex1.parent := tab;
    hex1.align := alClient;
    hex2 := nil;
  end else if mode=spCompare then begin
    CreatePanel(panel,ed1,alClient,panel.width div 2,0);
    CreatePanel(panel,ed2,alLeft,panel.width div 2,1);
    ed1.tag := 0;
    ed2.tag := 1;
    hex1 := TXSHexView.Create(tab);
    hex1.Tag:= 0;
    hex1.parent := tab;
    hex1.align := alClient;
    hex1.width := panel.width div 2;
    hex2 := TXSHexView.Create(tab);
    hex2.Tag:= 1;
    hex2.parent := tab;
    hex2.width := panel.width div 2;
    hex2.align := alLeft;

    panel := TPanel.Create(tab);
    panel.parent := tab;
    panel.align := alBottom;
    panel.Height:= 35;
    panel.color := cl3DLight;
    panel.BevelOuter:=bvNone;
    panel.BorderStyle:=bsNone;
    panel.OnResize:=@ComparePanelResize;

    button:= TButton.Create(panel);
    button.parent := panel;
    button.caption := 'Compare';
    button.top := 5;
    button.left:= (panel.width - button.width) div 2;
    button.onclick := @CompareClick;
  end;

  UpdateText;
  PageChange(self);
end;

procedure TXSDocManager.NewProject;
var
  i   : integer;
  n   : TTreeNode;
  s   : string;
  data: PXSTreeData;

  function SameCaption(AName: string): boolean;
  var
    i: integer;
  begin
    result := false;
    i:=0;
    while (i<ftreeview.Items.Count) and not result do begin
      if ftreeview.Items.Item[i].Parent=nil then
        result := ExtractFileName(ftreeview.Items.Item[i].Text) = AName;
      inc(i);
    end;
  end;

begin

  i := 1;
  s := fprjname+'1'+fprjext;
  while (i<100) and SameCaption(s) do begin
    inc(i);
    s := fprjname+IntToStr(i)+fprjext;
  end;

  NEW(data);
  data^.modified := false;
  data^.path := '';
  n := ftreeview.Items.AddObject(nil,s,data);
  n.ImageIndex    := 46;
  n.SelectedIndex := 46;
  n.StateIndex    := 46;
  if ftreeview.ActivePrj = nil then begin
    ftreeview.ActivePrj := n;
    ftreeview.ActivePrj.Selected := true;
  end;

end;

procedure TXSDocManager.SaveXmlProject(node: TTreeNode; name: string);
var
  Doc: TXMLDocument;
  RootNode, ItemNode, TextNode: TDOMNode;
  i: integer;

  procedure CreateSingle(var Element: TDOMNode; name,value: string);
  begin
    ItemNode:=Doc.CreateElement(name);
    TextNode:=Doc.CreateTextNode(value);
    ItemNode.AppendChild(TextNode);
    Element.AppendChild(ItemNode);
  end;

begin
  try
    Doc := TXMLDocument.Create;
    RootNode := Doc.CreateElement('Project');
    Doc.Appendchild(RootNode);

    CreateSingle(RootNode,'name',name);
    for i:=0 to node.Count-1 do begin
      CreateSingle(RootNode,'file',node.Items[i].Text);
      if node.Items[i].ImageIndex = 48 then TDOMElement(ItemNode).SetAttribute('main','true');
    end;

    WriteXMLFile(Doc,name);
  finally
    Doc.Free;
  end;
end;

procedure TXSDocManager.OpenXmlProject(node: TTreeNode; name: string);
var
  iNode, cNode: TDOMNode;
  Doc: TXMLDocument;
  n,a: TTreeNode;
begin
  if FileExists(name) then begin
    ReadXMLFile(Doc,name);
    iNode := Doc.DocumentElement.FirstChild;
    if iNode.NodeName='name' then begin
      node.text := iNode.FirstChild.NodeValue;
      PXSTreeData(node.data)^.path := name;
    end;
    iNode := iNode.NextSibling;
    a := nil;
    while iNode<>nil do begin
      if iNode.NodeName='file' then begin
        n := AddModInPrj(iNode.FirstChild.NodeValue);
        if iNode.HasAttributes and (iNode.Attributes.Length>0) then begin
          n.ImageIndex    := 48;
          n.SelectedIndex := 48;
          n.StateIndex    := 48;
          OpenPageByName(n.text);
          AddDebugTab(n.text);
          PageChange(self);
          a := n;
        end;
      end;
      iNode := iNode.NextSibling;
    end;
    if a<>nil then a.Selected:=true;
    Doc.Free;
  end;
end;

procedure TXSDocManager.OpenProject;
var
  n   : TTreeNode;
  data: PXSTreeData;
  i   : integer;
begin
  fopendialog.Filter := ffilter;
  fopendialog.FilterIndex := 2;
  fopendialog.FileName:='';
  if fopendialog.Execute then begin
    for i:=0 to fopendialog.Files.count-1 do begin
      if ftreeview.Items.FindNodeWithText(fopendialog.Files[i])=nil then begin
        NEW(data);
        data^.modified := false;
        data^.path := '';
        n := ftreeview.Items.AddObject(nil,'',data);
        n.ImageIndex    := 46;
        n.SelectedIndex := 46;
        n.StateIndex    := 46;
        ftreeview.ActivePrj := n;
        ftreeview.ActivePrj.Selected := true;
        OpenXmlProject(n,fopendialog.Files[i]);
        PXSTreeData(ftreeview.ActivePrj.Data)^.modified := false;
      end;
    end;
  end;
end;

function TXSDocManager.SaveProjectNode(node: TTreeNode): Boolean;
begin

  fsavedialog.Filter := ffilter;
  fsavedialog.FilterIndex := 2;
  fsavedialog.FileName := ExtractFileName(node.text);
  Result := fsavedialog.Execute;

  if Result then begin

    if FileExists(fsavedialog.FileName) then
      case MessageDlg('A file "' + fsavedialog.FileName + '" already exists.'+#13+'Replace it?' , mtWarning, mbYesNoCancel, 0) of
        mrYes:      Result := True;
        mrNo:       Result := False;
        mrCancel:   Result := False;
      end;

    if Result and (ExtractFilePath(node.text) <> '') and (node.text<>fsavedialog.FileName) and FileExists(node.text) then
      case MessageDlg('Delete old file "' + node.text + '" ?' , mtWarning, mbYesNoCancel, 0) of
        mrYes: DeleteFile(node.text);
      end;

    if Result then begin
      SaveXmlProject(node,fsavedialog.FileName);
      node.text := fsavedialog.FileName;
      PXSTreeData(node.data)^.Path     := fsavedialog.FileName;
      PXSTreeData(node.data)^.Modified := false;
      ftreeview.Invalidate;
    end;

  end;

end;

procedure TXSDocManager.SaveProjectAs;
var
  node: TTreeNode;
begin
  node := ftreeview.ActivePrj;
  if node<>nil then SaveProjectNode(node)
end;

function TXSDocManager.SaveProjectByNode(node: TTreeNode): boolean;
begin
  if node<>nil then begin
    Result := PXSTreeData(node.data)^.path <> '';
    if not Result then
      Result := SaveProjectNode(node)
    else if PXSTreeData(node.data)^.modified then begin
      SaveXmlProject(node,node.text);
      PXSTreeData(node.data)^.modified := false;
      ftreeview.Invalidate;
    end;
  end;
end;

procedure TXSDocManager.SaveProject;
begin
  SaveProjectByNode(ftreeview.ActivePrj);
end;

function TXSDocManager.CloseProjectNode(node: TTreeNode): Boolean;
var
  i: integer;
begin
  Result := not PXSTreeData(node.data)^.modified;
  if not Result then
    case MessageDlg('Save changed project "' + ExtractFileName(node.text) + '" ?' , mtWarning, mbYesNoCancel, 0) of
      mrYes:      Result := SaveProjectByNode(node);
      mrNo:       Result := True;
      mrCancel:   Result := False;
    end;
  if result then begin
    for i := node.count -1 downto 0 do begin
      if node.Items[i].Data<>nil then Dispose(PXSTreeData(node.Items[i].Data));
      ftreeview.Items.Delete(node.Items[i]);
    end;
    if node.Data<>nil then Dispose(PXSTreeData(node.Data));
    ftreeview.Items.Delete(node);
    if ftreeview.Items.count > 0 then
      ftreeview.ActivePrj := ftreeview.Items.Item[0]
    else
      ftreeview.ActivePrj := nil;
  end;
  ftreeview.ActiveNode := nil;
end;

procedure TXSDocManager.CloseProject;
begin
  if ftreeview.ActiveNode<>nil then begin
    CloseProjectNode(ftreeview.ActiveNode)
  end else if ftreeview.ActivePrj<>nil then begin
    CloseProjectNode(ftreeview.ActivePrj)
  end;
end;

procedure TXSDocManager.SaveProjectFromMenu;
begin
  if ftreeview.ActiveNode<>nil then begin
    SaveProjectByNode(ftreeview.ActiveNode)
  end else if ftreeview.ActivePrj<>nil then begin
    SaveProjectByNode(ftreeview.ActivePrj)
  end;
end;

procedure TXSDocManager.SaveProjectAsFromMenu;
begin
  if ftreeview.ActiveNode<>nil then begin
    SaveProjectNode(ftreeview.ActiveNode)
  end else if ftreeview.ActivePrj<>nil then begin
    SaveProjectNode(ftreeview.ActivePrj)
  end;
end;

procedure TXSDocManager.SetAsMainFromMenu;
var
  node: TTreeNode;
  i,j : integer;
begin
  if ftreeview.ActiveNode<>nil then
    node := ftreeview.ActiveNode
  else if ftreeview.ActivePrj<>nil then
    node := ftreeview.ActivePrj;
  if (node <> nil) and (ftreeview.ActiveChild<>nil) and (ftreeview.ActiveChild.ImageIndex <> 48) then begin
    i := 0;
    while (i < node.Count) do begin
      if node.Items[i].ImageIndex = 48 then begin
        node.Items[i].ImageIndex    := 47;
        node.Items[i].SelectedIndex := 47;
        node.Items[i].StateIndex    := 47;
        j := FindPageByName(node.Items[i].text);
        if j>-1 then fpagecontrol.DelTabByIdx(j,1);
        break;
      end;
      inc(i);
    end;
    PXSTreeData(node.data)^.modified := true;
    ftreeview.ActiveChild.ImageIndex    := 48;
    ftreeview.ActiveChild.SelectedIndex := 48;
    ftreeview.ActiveChild.StateIndex    := 48;
    AddDebugTab(ftreeview.ActiveChild.text);
    fpagecontrol.Invalidate;
  end;
end;

function TXSDocManager.PrjIsModified(flag: boolean): boolean;
begin
  result := false;
  if not flag and (ftreeview.ActiveNode<>nil) then
    result := PXSTreeData(ftreeview.ActiveNode.data)^.Modified or (PXSTreeData(ftreeview.ActiveNode.data)^.path='')
  else if ftreeview.ActivePrj<>nil then
    result := PXSTreeData(ftreeview.ActivePrj.data)^.Modified or (PXSTreeData(ftreeview.ActivePrj.data)^.path='');
end;

procedure TXSDocManager.AddFilesInPrj;
var
  node: TTreeNode;
  i   : integer;
begin
  if ftreeview.ActiveNode<>nil then
    node := ftreeview.ActiveNode
  else if ftreeview.ActivePrj<>nil then
    node := ftreeview.ActivePrj;
  if node <> nil then begin
    fopendialog.Filter := ffilter;
    fopendialog.FilterIndex := 0;
    fopendialog.FileName:='';
    if fopendialog.Execute then begin
      for i:=0 to fopendialog.Files.count-1 do begin
        Application.ProcessMessages;
        AddModInPrjByNode(node,fopendialog.Files[i]);
      end;
      ftreeview.Invalidate;
    end;
  end;
end;

procedure TXSDocManager.RemoveFileInPrj;
var
  node: TTreeNode;
begin
  if ftreeview.ActiveChild<>nil then begin
    node := ftreeview.ActiveChild.parent;
    if node <> nil then begin
      if ftreeview.ActiveChild.data<>nil then Dispose(PXSTreeData(ftreeview.ActiveChild.data));
      ftreeview.Items.Delete(ftreeview.ActiveChild);
      ftreeview.ActiveChild := nil;
      PXSTreeData(node.data)^.Modified := true;
    end;
  end;
end;

function TXSDocManager.SaveCloseQueryProjects: Boolean;
var
  i   : integer;
  node: TTreeNode;
begin
  Result:=False;
  for i := ftreeview.items.count-1 downto 0 do begin
    node := ftreeview.items.item[i];
    if (node.parent=nil) and PXSTreeData(node.data)^.modified then
      case MessageDlg('Save changed project "' + ExtractFileName(node.text) + '" ?' , mtWarning, mbYesNoCancel, 0) of
        mrYes   : if not SaveProjectByNode(node) then Exit;
        mrNo    : ;
        mrCancel: begin Result := False; Exit; end;
      end;
  end;
  Result:=True;
end;

procedure TXSDocManager.AddDebugTab(name: string);
var
  tab : TXSScrollbox;
  i   : integer;
begin
  i := FindPageByName(name);
  if i>-1 then begin
    tab := fpagecontrol.AddTabInIdx(i,'Debug','');
    tab.VScroll.Visible:=false;
    tab.HScroll.Visible:=false;
  end;
end;

procedure TXSDocManager.TreeDblClick(Sender: TObject);
var
  node: TTreeNode;
  tab : TXSScrollbox;
  i   : integer;
begin
  node := ftreeview.Selected;
  if (node<>nil) then begin
    if node.parent=nil then begin
      ftreeview.ActivePrj := node;
      ftreeview.ActivePrj.Expanded := true;
    end else begin
      OpenPageByName(node.text);
      if node.ImageIndex = 48 then AddDebugTab(node.text);
      PageChange(self);
    end;
  end;
end;

procedure TXSDocManager.TreeOnClick(Sender: TObject);
var
  node: TTreeNode;
  i   : integer;
  f   : boolean;
  tab : TXSScrollbox;
begin
  node := ftreeview.Selected;
  if (node<>nil) then begin
    if node.parent<>nil then begin
      i := 0; f := false;
      while (i<fpagecontrol.pagelist.count) and not f do begin
        tab := GetTabByNum(i,0);
        f := tab.path = node.text;
        inc(i);
      end;
      if f then begin
        OpenPageByName(node.text);
        PageChange(self);
      end;
    end;
  end;
end;

function TXSDocManager.AddModInPrjByNode(node: TTreeNode; path: string): TTreeNode;
var
  data: PXSTreeData;
begin
  result := nil;
  if (node<>nil) then begin
    if node.FindNode(path)=nil then begin
      PXSTreeData(node.data)^.modified := true;
      NEW(data);
      data^.modified := false;
      data^.path := '';
      PXSTreeData(node.Data)^.modified := true;
      node := ftreeview.Items.AddChildObject(node,path,data);
      node.ImageIndex    := 47;
      node.SelectedIndex := 47;
      node.StateIndex    := 47;
      result := node;
      SelectNode(path);
    end;
  end;
end;

function TXSDocManager.AddModInPrj(path: string): TTreeNode;
begin
  result := AddModInPrjByNode(ftreeview.ActivePrj,path);
end;

procedure TXSDocManager.SelectNode(path: string);
var
  node: TTreeNode;
begin
  if (ftreeview.Items.Count>0) then begin
    if (path='') and (ftreeview.ActivePrj<>nil) then begin
      ftreeview.ActivePrj.selected := true;
      exit;
    end;
    node := ftreeview.Items.FindNodeWithText(path);
    if node<>nil then node.selected := true;
  end;
end;

procedure TXSDocManager.PopupClosePage;
begin
  if fpagecontrol.popupidx<>-1 then begin
    CloseFileByIdx(fpagecontrol.popupidx);
    fpagecontrol.Invalidate;
  end;
end;

procedure TXSDocManager.PopupAddInPrj;
var
  tab: TXSScrollbox;
  node: TTreeNode;
begin
  if fpagecontrol.popupidx<>-1 then begin
    tab := GetTabByNum(fpagecontrol.popupidx,0);
    if (tab<>nil) and (ftreeview.ActivePrj<>nil) then begin
      AddModInPrj(tab.path);
      if fpagecontrol.ActivePage = fpagecontrol.popupidx then ;
      ftreeview.Invalidate;
    end;
  end;
end;

procedure TXSDocManager.PopupSetAsMain;
var
  tab : TXSScrollbox;
  node,root: TTreeNode;
  i,j : integer;
  f   : boolean;
begin
  if fpagecontrol.popupidx<>-1 then begin
    tab := GetTabByNum(fpagecontrol.popupidx,0);
    if tab<>nil then begin
      root := ftreeview.Items.FindNodeWithText(tab.path);
      if root<>nil then root := root.parent;
      if root<>nil then begin
        node := root.FindNode(tab.path);
        if (node <> nil) and (node.ImageIndex <> 48) then begin
          i := 0;
          while (i < root.Count) do begin
            if root.Items[i].ImageIndex = 48 then begin
              root.Items[i].ImageIndex    := 47;
              root.Items[i].SelectedIndex := 47;
              root.Items[i].StateIndex    := 47;
              j := FindPageByName(root.Items[i].text);
              if j>-1 then fpagecontrol.DelTabByIdx(j,1);
              break;
            end;
            inc(i);
          end;
          PXSTreeData(root.data)^.modified := true;
          node.ImageIndex    := 48;
          node.SelectedIndex := 48;
          node.StateIndex    := 48;
          AddDebugTab(node.text);
          fpagecontrol.Invalidate;
        end;
      end;
      ftreeview.Invalidate;
    end;
  end;
end;

function TXSDocManager.PageTemplate(name: string): string;
begin
  result :=     'MODULE '+name+';'+#10#13;
  result := result + #10#13;
  result := result + 'IMPORT'+#10#13;
  result := result + '   SYSTEM;'+#10#13;
  result := result + #10#13;
  result := result + 'BEGIN'+#10#13;
  result := result + #10#13;
  result := result + 'END '+name+'.';
end;

procedure TXSDocManager.NewPage;
var
  i  : integer;
  s  : string;
  tab: TXSScrollbox;
  ed : TXSSynEdit;

  function SameCaption(AName: string): boolean;
  var
    i: integer;
  begin
    result := false;
    i:=0;
    while (i<fpagecontrol.pagelist.count) and not result do begin
      result := fpagecontrol.pagelist[i] = AName;
      inc(i);
    end;
  end;

begin
  i := 1;
  s := fdefname+'1'+fdefext;
  while (i<100) and SameCaption(s) do begin
    inc(i);
    s :=fdefname+IntToStr(i)+fdefext;
  end;

  tab := fpagecontrol.AddPage(s,s,'Source',false);

  ed            := TXSSynEdit.Create(Form1);
  ed.encode     := 'UTF8';
  ed.Visible    := false;
  EditorOptions(ed,tab);
  ed.Text := PageTemplate(fdefname+IntToStr(i));
  ed.Visible    := true;

  PageChange(self);
end;

function TXSDocManager.FindPageByName(name: string): integer;
var
  j     : integer;
  tab   : TXSScrollBox;
begin
  Result := -1;
  tab := nil; j := 0;
  while (j<=fpagecontrol.pagelist.count-1) do begin
    tab := GetTabByNum(j,0);
    if (tab.Path = name) then break else tab := nil;
    inc(j);
  end;
  if (tab<>NIL) and not tab.Special then Result := j;
end;

procedure TXSDocManager.OpenPageByName(name: string);
var
  ed    : TXSSynEdit;
  j     : integer;
  fn,fp : string;
  s     : string;
  tab   : TXSScrollBox;
  sl    : TStringList;
begin

  fn := ExtractFileName(name);
  fp := ExtractFilePath(name);

  tab := nil; j := 0;
  while (j<=fpagecontrol.pagelist.count-1) do begin
    tab := GetTabByNum(j,0);
    if (tab.Path = fp+fn) then break else tab := nil;
    inc(j);
  end;

  if tab <> NIL then begin
    fpagecontrol.ActivePage := j;
    PageChange(self);
  end else begin
    tab := fpagecontrol.AddPage(fn,fp+fn,'Code',true);

    ed            := TXSSynEdit.Create(Form1);
    ed.Visible    := false;
    EditorOptions(ed,tab);

    if fp = '' then begin
      s := '';
      j := 1;
      while (j < length(fn)) and (fn[j]<>'.') do begin
        s := s + fn[j];
        inc(j);
      end;
      ed.Text := PageTemplate(s);
      ed.encode := 'UTF8';
    end else begin
      sl := TStringList.Create;
      sl.LoadFromFile(fp+fn);
      ed.encode := LConvEncoding.GuessEncoding(sl.Text);
      ed.Text := LConvEncoding.ConvertEncoding(sl.Text, ed.encode, 'UTF-8');
      sl.Free;
    end;

    ed.Modified   := false;
    //ed.Visible    := true;
  end;

end;

procedure TXSDocManager.OpenPage;
var
  i   : integer;
begin
  fopendialog.Filter := ffilter;
  fopendialog.FilterIndex := 0;
  fopendialog.FileName:='';
  if fopendialog.Execute then begin
    for i:=0 to fopendialog.Files.count-1 do begin
      Application.ProcessMessages;
      OpenPageByName(fopendialog.Files[i]);
    end;
    PageChange(self);
  end;
end;

procedure TXSDocManager.ClosePage;
begin
  CloseFileByIdx(fpagecontrol.activepage);
  fpagecontrol.Invalidate;
end;

procedure TXSDocManager.SaveText(idx: integer; path, from, encode: string);
var
  ed   : TXSSynEdit;
  sl   : TStringList;
begin
  ed := GetEditorByNum(idx,0);
  if from<>encode then begin
    sl := TStringList.Create;
    sl.Text := LConvEncoding.ConvertEncoding(ed.Text, from, encode);
    sl.SaveToFile(path);
    sl.Free;
  end else
    ed.Lines.SaveToFile(path);
  ed.Modified := false;
end;

function TXSDocManager.SaveFileAsByIdx(idx: integer): Boolean;
var
  tab  : TXSScrollbox;
  node : TTreeNode;
begin

  tab   := GetTabByNum(idx,0);
  if tab = nil then exit;
  if tab.Special then exit;

  fsavedialog.Filter := ffilter;
  fsavedialog.FilterIndex := 0;
  fsavedialog.FileName := fpagecontrol.pagelist[idx];
  Result := fsavedialog.Execute;

  if Result then begin

    if FileExists(fsavedialog.FileName) and (tab.Path<>fsavedialog.FileName) then
      case MessageDlg('A file "' + fsavedialog.FileName + '" already exists.'+#13+'Replace it?' , mtWarning, mbYesNoCancel, 0) of
        mrYes:      Result := True;
        mrNo:       Result := False;
        mrCancel:   Result := False;
      end;

    if Result and (ExtractFilePath(tab.Path) <> '') and (tab.Path<>fsavedialog.FileName) and FileExists(tab.Path) then
      case MessageDlg('Delete old file "' + tab.Path + '" ?' , mtWarning, mbYesNoCancel, 0) of
        mrYes: DeleteFile(tab.Path);
      end;

    if Result then begin
      if ExtractFileExt(fsavedialog.FileName) = '' then fsavedialog.FileName:=fsavedialog.FileName+fdefext;
      SaveText(idx,fsavedialog.FileName,'UTF-8', GetEditorByNum(idx,0).encode);
      fpagecontrol.pagelist[idx] := ExtractFileName(fsavedialog.FileName);

      node := ftreeview.Items.FindNodeWithText(tab.Path);
      if node <> nil then begin
        node.text := fsavedialog.FileName;
        if node.parent<>nil then PXSTreeData(node.parent.data)^.modified := true;
      end;

      tab.Path    := fsavedialog.FileName;
      tab.Modified := false;
      UpdateText;
      InvalidateParent(fpagecontrol.Parent);
    end;

  end;

end;

procedure TXSDocManager.SaveFileAs;
begin
  SaveFileAsByIdx(fpagecontrol.activepage);
end;

function TXSDocManager.SaveFileByIdx(idx: integer): Boolean;
var
  tab  : TXSScrollbox;
begin
  tab    := GetTabByNum(idx,0);
  Result := not tab.Special and (ExtractFilePath(tab.Path) <> '');
  if not Result then
    Result := SaveFileAsByIdx(idx)
  else if tab.Modified then begin
    SaveText(idx,tab.path,'UTF-8', GetEditorByNum(idx,0).encode);
    tab.Path     := tab.path;
    tab.Modified := false;
    UpdateText;
    InvalidateParent(fpagecontrol.parent);
  end;
end;

procedure TXSDocManager.SaveFile;
begin
  SaveFileByIdx(fpagecontrol.activepage);
end;

procedure TXSDocManager.SaveAll;
var
  i: integer;
begin
  for i:=fpagecontrol.pagelist.count-1 downto 0 do
    if not SaveFileByIdx(i) then exit;
end;

function TXSDocManager.CloseFileByIdx(idx: integer): Boolean;
var
  tab  : TXSScrollbox;
begin
  tab := GetTabByNum(idx,0);
  Result := tab.Special or not tab.Modified;
  if not Result then
    case MessageDlg('Save changed file "' + fpagecontrol.pagelist[idx] + '" ?' , mtWarning, mbYesNoCancel, 0) of
      mrYes:      Result := SaveFileByIdx(idx);
      mrNo:       Result := True;
      mrCancel:   Result := False;
    end;
  if result then fpagecontrol.ClosePage(idx);
end;

procedure TXSDocManager.CloseFile;
begin
  CloseFileByIdx(fpagecontrol.activepage);
end;

function TXSDocManager.CloseAll: Boolean;
begin
  Result := False;
  while fpagecontrol.pagelist.count > 0 do
    if not CloseFileByIdx(fpagecontrol.pagelist.count-1) then Exit;
  Result := True;
end;

function TXSDocManager.SaveCloseQuery: Boolean;
var
  i: integer;
  tab  : TXSScrollbox;
begin
  Result:=False;
  for i:=fpagecontrol.pagelist.count-1 downto 0 do begin
    tab := GetTabByNum(i,0);
    if not tab.Special and tab.Modified then
      case MessageDlg('Save changed file "' + fpagecontrol.pagelist[i] + '" ?' , mtWarning, mbYesNoCancel, 0) of
        mrYes   : if not SaveFileByIdx(i) then Exit;
        mrNo    : ;
        mrCancel: begin Result := False; Exit; end;
      end;
  end;
  Result:=True;
end;

procedure TXSDocManager.Cut;
begin
  if feditor = nil then exit;
  feditor.CutToClipboard;
end;

procedure TXSDocManager.Copy;
begin
  if feditor = nil then exit;
  feditor.CopyToClipboard;
end;

procedure TXSDocManager.Paste;
begin
  if feditor = nil then exit;
  feditor.PasteFromClipboard;
end;

procedure TXSDocManager.Delete;
begin
  if feditor = nil then exit;
  feditor.ClearSelection;
end;

procedure TXSDocManager.Undo;
begin
  if feditor = nil then exit;
  feditor.Undo;
end;

procedure TXSDocManager.Redo;
begin
  if feditor = nil then exit;
  feditor.Redo;
end;

procedure TXSDocManager.Find;
var
  key: Word;
begin
  if feditor = nil then exit;
  if not Form1.findpanel.Visible then Form1.findpanel.Visible:=true;
  foldtext := Form1.Edit1.Text;
  if not SameText(Form1.Edit1.Text,feditor.SelText) and (feditor.SelText<>'') then
    Form1.Edit1.Text:=feditor.SelText;
  feditor.SetFocus;
  key := VK_F3;
  if not fdown then EditorKeyDown(feditor,key,[]);
  fdown := false;
end;

procedure TXSDocManager.FindNext;
begin
  if feditor = nil then exit;
  if Form1.CheckBox1.Checked then
    fsearchopt := fsearchopt + [ssoMatchCase]
  else
    fsearchopt := fsearchopt - [ssoMatchCase];

  if Form1.CheckBox2.Checked then
    fsearchopt := fsearchopt + [ssoWholeWord]
  else
    fsearchopt := fsearchopt - [ssoWholeWord];

  fsearchopt := fsearchopt-[ssoBackwards];
  Form1.RadioButton2.Checked:=true;
  if foldtext = Form1.Edit1.Text then SearchReplace;
  foldtext := Form1.Edit1.Text;
end;

procedure TXSDocManager.FindPrev;
begin
  if feditor = nil then exit;
  if Form1.CheckBox1.Checked then
    fsearchopt := fsearchopt + [ssoMatchCase]
  else
    fsearchopt := fsearchopt - [ssoMatchCase];

  if Form1.CheckBox2.Checked then
      fsearchopt := fsearchopt + [ssoWholeWord]
    else
      fsearchopt := fsearchopt - [ssoWholeWord];

  fsearchopt:=fsearchopt+[ssoBackwards];
  Form1.RadioButton1.Checked:=true;
  if foldtext = Form1.Edit1.Text then SearchReplace;
  foldtext := Form1.Edit1.Text;
end;

procedure TXSDocManager.Unindent;
begin
  if feditor = nil then exit;
  feditor.CommandProcessor(ecBlockUnIndent,#0,nil);
end;

procedure TXSDocManager.Indent;
begin
  if feditor = nil then exit;
  feditor.CommandProcessor(ecBlockIndent,#0,nil);
end;

procedure TXSDocManager.Build;
begin
end;

function TXSDocManager.Run: boolean;
begin
  //Parser.Compile('',@Editor.Text[1],0);
end;

procedure TXSDocManager.Pause;
begin
end;

procedure TXSDocManager.Stop;
begin
end;

procedure TXSDocManager.SelectAll;
begin
  if feditor = nil then exit;
  feditor.SelectAll;
end;

procedure TXSDocManager.DeleteSelection;
begin
  if feditor = nil then exit;
  feditor.CommandProcessor(ecLeft,#0,nil);
  feditor.CommandProcessor(ecRight,#0,nil);
end;

procedure TXSDocManager.Upper;
var
  x,y,c: TPoint;
begin
  if feditor = nil then exit;
  x := feditor.BlockBegin;
  y := feditor.BlockEnd;
  c := feditor.CaretXY;
  feditor.SelText := UpperCase(feditor.SelText);
  feditor.BlockBegin := x;
  feditor.BlockEnd := y;
  feditor.CaretXY := c;
end;

procedure TXSDocManager.Lower;
var
  x,y,c: TPoint;
begin
  if feditor = nil then exit;
  x := feditor.BlockBegin;
  y := feditor.BlockEnd;
  c := feditor.CaretXY;
  feditor.SelText := LowerCase(feditor.SelText);
  feditor.BlockBegin := x;
  feditor.BlockEnd := y;
  feditor.CaretXY := c;
end;

procedure TXSDocManager.Swap;
var
  s: string;
  i: integer;
  x,y,c: TPoint;
begin
  if feditor = nil then exit;
  x := feditor.BlockBegin;
  y := feditor.BlockEnd;
  c := feditor.CaretXY;
  s := feditor.SelText;
  i:=1;
  while i<=length(s) do begin
    if upCase(s[i])=s[i] then
      s[i]:=LowerCase(s[i])
    else
      s[i]:=upCase(s[i]);
    inc(i);
  end;
  feditor.SelText := s;
  feditor.BlockBegin := x;
  feditor.BlockEnd := y;
  feditor.CaretXY := c;
end;

procedure TXSDocManager.TitleUp;
var
  s: string;
  i: integer;
  x,y,c: TPoint;
begin
  if feditor = nil then exit;
  x := feditor.BlockBegin;
  y := feditor.BlockEnd;
  c := feditor.CaretXY;
  s := feditor.SelText;
  i:=1;
  while i<=length(s) do begin
    if (i>1) and not (s[i-1] in ['a'..'z','A'..'Z']) then s[i]:=upCase(s[i]);
    if (i=1) and (s[i] in ['a'..'z','A'..'Z'])       then s[i]:=upCase(s[i]);
    inc(i);
  end;
  feditor.SelText := s;
  feditor.BlockBegin := x;
  feditor.BlockEnd := y;
  feditor.CaretXY := c;
end;

procedure TXSDocManager.TitleLo;
var
  s: string;
  i: integer;
  x,y,c: TPoint;
begin
  if feditor = nil then exit;
  x := feditor.BlockBegin;
  y := feditor.BlockEnd;
  c := feditor.CaretXY;
  s := feditor.SelText;
  i:=1;
  while i<=length(s) do begin
    if (i>1) and not (s[i-1] in ['a'..'z','A'..'Z']) then s[i]:=LowerCase(s[i]);
    if (i=1) and (s[i] in ['a'..'z','A'..'Z'])       then s[i]:=LowerCase(s[i]);
    inc(i);
  end;
  feditor.SelText := s;
  feditor.BlockBegin := x;
  feditor.BlockEnd := y;
  feditor.CaretXY := c;
end;

procedure TXSDocManager.Browse;
begin
end;

end.

