unit XsEditors;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Controls, ExtCtrls, StdCtrls, Messages, LMessages, XsControls,
  Types, LCLType, LCLintf, Graphics, Dialogs;

type

  { TXSHexView }

  TCursorPos = record
    hex   : TPoint;
    text  : TPoint;
    frame : TPoint;
    endh  : TPoint;
    endt  : TPoint;
  end;

  TData = array of byte;

  TXSHexView = class(TXSScrollbox)
  private
    fdata      : TData;
    fblink     : boolean;
    fblinkclr  : integer;
    floaded    : boolean;
    fpix       : integer;
    flen       : integer;
    fadr       : pointer;
    fstart     : integer;
    fhexpix    : integer;
    fborder    : integer;
    fdown      : boolean;
    fhexarea   : TRect;
    ftextarea  : TRect;
    fheadarea  : TRect;
    fgutterarea: TRect;
    ftextpos   : TPoint;
    fhexactive : boolean;
    ftextactive: boolean;
    fcharscount: integer;
    fcharheight: integer;
    fcharwidth : integer;
    fcurstart  : integer;
    fcurend    : integer;
    fmousex    : integer;
    cursorpos  : TCursorPos;
    fgutterwidth : integer;
    fvisibleline : integer;
    ftimer     : TTimer;
    frframe    : boolean;
    procedure   UpdateArea;
    procedure   UpdateSelection(X,Y: integer; flag: boolean);
    procedure   UpdateCursor;
    procedure   TimerTime(Sender: TObject);
    procedure   Swap(var x,y: integer);
  protected
    procedure   Paint; override;
    procedure   Resize; override;
    procedure   MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure   MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure   WMMouseWheel(var Message: TLMMouseEvent); message LM_MOUSEWHEEL;
    procedure   KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure   ScrollChange(Sender: TObject);
    procedure   WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
  public
    ChangedArray : array of boolean;
    constructor Create(TheOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   LoadFromFile(const FileName: string);
    procedure   CopyToClipboard;
  published
    property    Len : integer read flen;
    property    Data: TData read fdata;
  end;

implementation

{ TXSHexView }

constructor TXSHexView.Create(TheOwner: TComponent);
var
  i: integer;
begin
  inherited Create(TheOwner);
  HScroll.visible := false;
  VScroll.OnChange := @ScrollChange;
  Font.Name := 'Courier New';
  Font.Size := 9;
  fhexarea   := Rect(0,0,0,0);
  ftextarea  := Rect(0,0,0,0);
  fheadarea  := Rect(0,0,0,0);
  fgutterarea:= Rect(0,0,0,0);
  fborder    := 5;
  fpix       := 3;
  fstart     := 0;
  floaded    := true;
  fvisibleline := 0;
  fhexactive := true;
  ftextactive:= false;
  fdown      := false;
  fcurstart  := 0;
  fcurend    := 0;
  fmousex    := 0;
  ftimer          := TTimer.Create(self);
  ftimer.Interval := 500;
  ftimer.OnTimer  := @TimerTime;
  ftimer.Enabled  := true;
  fblinkclr       := $E6B7B7;
  frframe         := false;

  flen := $200;
  SetLength(fdata,flen+1);
  fillchar(fdata[0],flen,#0);
  SetLength(ChangedArray,flen+1);
  for i:=0 to flen do ChangedArray[i]:=false;
end;

destructor TXSHexView.Destroy;
begin
  SetLength(fdata,0);
  SetLength(ChangedArray,0);
  ftimer.free;
  inherited Destroy;
end;

procedure TXSHexView.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  inherited;
  Msg.Result := DLGC_WANTARROWS or DLGC_WANTCHARS or DLGC_WANTALLKEYS;
  //if fWantTabs and (GetKeyState(VK_CONTROL) >= 0) then
    //Msg.Result := Msg.Result or DLGC_WANTTAB;
end;

procedure TXSHexView.TimerTime(Sender: TObject);
var
  r: PRect;
begin
  fblink := not fblink;
  if fblink then
    fblinkclr := $C88C8C
  else
    fblinkclr := $E6B7B7;

  new(r);
  r^ := Rect(cursorpos.frame.x,cursorpos.frame.y,cursorpos.frame.x+fcharwidth,cursorpos.frame.y+fcharheight);
  InvalidateRect(handle,r,false);
  r^ := Rect(cursorpos.text.x,cursorpos.text.y,cursorpos.text.x+fcharwidth,cursorpos.text.y+fcharheight);
  InvalidateRect(handle,r,false);
  DISPOSE(r);
end;

procedure TXSHexView.LoadFromFile(const FileName: string);
var
  f: TFileStream;
  i: integer;
begin
  fdown      := false;
  fcurstart  := 0;
  fcurend    := 0;
  fmousex    := 0;
  fstart := 0;
  flen := 0;
  SetLength(fdata,flen);
  SetLength(ChangedArray,flen);
  VScroll.position := 0;
  UpdateArea;

  f := TFileStream.Create(FileName,fmOpenRead);
  f.Seek(0,soBeginning);
  flen := f.size;
  SetLength(fdata,flen);
  f.Read(fdata[0],flen);
  SetLength(ChangedArray,flen);
  for i:=0 to flen-1 do ChangedArray[i]:=false;
  f.free;
  SetFocus;
  UpdateCursor;
  Invalidate;
end;

procedure TXSHexView.Swap(var x,y: integer);
var
  i: integer;
begin
  x := fcurstart;
  y := fcurend;
  if x > y then begin
    i := x;
    x := y;
    y := i;
  end;
end;

procedure TXSHexView.CopyToClipboard;
var
  s: string;
  x,y: integer;
begin
  Swap(x,y);
  SetLength(s,abs(y-x)+1);
  Move(fdata[x],s[1],abs(y-x)+1);
  ShowMessage(s);
end;

procedure TXSHexView.WMMouseWheel(var Message: TLMMouseEvent);
begin
  if Message.WheelDelta > 0 then
    VScroll.position := VScroll.position - 4
  else
    VScroll.position := VScroll.position + 4;
  if VScroll.position < 0 then VScroll.position := 0;
  if VScroll.position > VScroll.Max-VScroll.PageSize then VScroll.position := VScroll.Max-VScroll.PageSize;
  fstart := VScroll.position;
  UpdateCursor;
  Invalidate;
end;

procedure TXSHexView.ScrollChange(Sender: TObject);
begin
  fstart := VScroll.position;
  UpdateCursor;
  Invalidate;
end;

procedure TXSHexView.UpdateSelection(X,Y: integer; flag: boolean);
var
  i: integer;
begin

  if flag then begin
    i := 0;
    ftextpos.x := fhexarea.left;
    while ftextpos.x<X do begin
      if (i>0) and (i mod 8 = 0) then inc(ftextpos.x,fcharwidth);
      inc(ftextpos.x,fhexpix);
      inc(i);
    end;
    ftextpos.x := i-1;

    if fdown then begin
      if (X<fhexarea.left)  then ftextpos.x := 0;
      if (X>fhexarea.right) then ftextpos.x := fcharscount-1;
    end;

  end else begin
    ftextpos.x := (X - ftextarea.left) div fcharwidth;
    if fdown then begin
      if (X<ftextarea.left)  then ftextpos.x := 0;
      if (X>ftextarea.right) then ftextpos.x := fcharscount-1;
    end;
    frframe := false;
  end;

  ftextpos.y := (Y - fhexarea.top)  div fcharheight;

  if ftextpos.x < 0 then ftextpos.x := 0;
  if (ftextpos.x > fcharscount-1) then ftextpos.x := fcharscount-1;

  if (fstart=0) and (ftextpos.y < 0) then
    ftextpos.y := 0
  else if (fstart>0) and (ftextpos.y < 0) then
    dec(fstart,4);

  if (ftextpos.y > fvisibleline-1) and (fstart > flen div fcharscount - fvisibleline) then
    ftextpos.y := fvisibleline -1
  else if (ftextpos.y > fvisibleline-1) and (fstart <= flen div fcharscount - fvisibleline) then
    inc(fstart,4);

  if fstart >= flen div fcharscount - fvisibleline + 1 then fstart := flen div fcharscount - fvisibleline + 1;
  fmousex := X;

  if (ftextpos.y > flen div fcharscount) then ftextpos.y := flen div fcharscount;
  if fvisibleline*fcharscount-(fcharscount-ftextpos.x)>flen then fstart := 0;

  VScroll.position := fstart;
end;

procedure TXSHexView.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  fdown := Button=mbLeft;
  if fdown then begin
    SetFocus;
    fhexactive  := PtInRect(fhexarea,Point(X,Y));
    ftextactive := PtInRect(ftextarea,Point(X,Y));
    fhexactive := not ftextactive;
    UpdateSelection(X,Y,fhexactive);
    fcurstart  := ftextpos.x + (ftextpos.y + fstart) * fcharscount;
    fcurend    := fcurstart;
    UpdateCursor;
    Invalidate;
  end;
end;

procedure TXSHexView.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if fdown then begin
    UpdateSelection(X,Y,fhexactive);
    fcurstart := ftextpos.x + (ftextpos.y + fstart) * fcharscount;
    UpdateCursor;
    Invalidate;
  end;
end;

procedure TXSHexView.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if fdown then begin
    fdown := false;
    UpdateSelection(X,Y,fhexactive);
    fcurstart := ftextpos.x + (ftextpos.y + fstart) * fcharscount;
    UpdateCursor;
    Invalidate;
  end;
end;

procedure TXSHexView.KeyDown(var Key: Word; Shift: TShiftState);
var
  i,x,y: integer;
  flag : boolean;
begin

  flag := (ssShift in Shift);

  if fhexactive and not flag and not (ssCtrl in Shift) then begin
    i := 0;
    fmousex := fhexarea.left;
    while i < fcurstart mod fcharscount do begin
      if (i+1) mod 8 = 0 then inc(fmousex,fcharwidth);
      inc(fmousex,fhexpix);
      inc(i);
    end;
    if frframe then inc(fmousex,fcharwidth+1);
  end;

  case Key of
    VK_CONTROL : flag := true;
    VK_SHIFT   : flag := true;
    ORD('C')   : if ssCtrl in Shift then begin flag := true; CopyToClipboard; end;
    VK_RIGHT   : if fhexactive and not frframe then inc(fmousex,fcharwidth+1) else inc(fcurstart);
    VK_LEFT    : if fhexactive and frframe then dec(fmousex,fcharwidth+1) else dec(fcurstart);
    VK_DOWN    : inc(fcurstart,fcharscount);
    VK_UP      : if fcurstart > fcharscount-1 then dec(fcurstart,fcharscount);
    VK_HOME    : dec(fcurstart,fcurstart mod fcharscount);
    VK_END     : inc(fcurstart,fcharscount - (fcurstart mod fcharscount) - 1);
    VK_PRIOR   : begin dec(fcurstart,fcharscount*fvisibleline); dec(fstart,fvisibleline); end;
    VK_NEXT    : begin inc(fcurstart,fcharscount*fvisibleline); inc(fstart,fvisibleline); end;
  end;

  if not flag then begin
    fcurend := fcurstart;
    y := fcurstart div fcharscount;
    x := fcurstart mod fcharscount;
    if (y < fstart) and (fstart > 0) then
      dec(fstart,1);
    if (y > fstart + fvisibleline-1) and (fstart <= flen div fcharscount - fvisibleline) then
      inc(fstart,1);
    if fstart < 0 then fstart := 0;
    if fstart >= flen div fcharscount - fvisibleline + 1 then fstart := flen div fcharscount - fvisibleline + 1;
    if (y > flen div fcharscount) then y := flen div fcharscount;
    if fvisibleline*fcharscount-(fcharscount-x)>flen then fstart := 0;
  end;

  UpdateCursor;
  VScroll.position := fstart;
  Invalidate;
end;

procedure TXSHexView.UpdateCursor;
var
  i    : integer;
begin
  if fcharscount <= 0 then exit;

  if fcurstart < 0      then fcurstart := 0;
  if fcurend   < 0      then fcurend   := 0;
  if flen > 0 then begin
    if fcurstart > flen-1 then fcurstart := flen-1;
    if fcurend   > flen-1 then fcurend   := flen-1;
  end;

  i := 0;
  cursorpos.hex.x := fhexarea.left;
  while i < fcurstart mod fcharscount do begin
    if (i+1) mod 8 = 0 then inc(cursorpos.hex.x,fcharwidth);
    inc(cursorpos.hex.x,fhexpix);
    inc(i);
  end;
  cursorpos.hex.y := fhexarea.top + (fcurstart div fcharscount - fstart)* fcharheight;

  cursorpos.frame.x := cursorpos.hex.x;
  cursorpos.frame.y := cursorpos.hex.y;
  frframe := (fmousex > cursorpos.frame.x+fcharwidth) and fhexactive;
  //if (fcurstart=fcurend) and
  if frframe then inc(cursorpos.frame.x,fcharwidth);

  i := 0;
  cursorpos.endh.x := fhexarea.left;
  while i < fcurend mod fcharscount do begin
    if (i+1) mod 8 = 0 then inc(cursorpos.endh.x,fcharwidth);
    inc(cursorpos.endh.x,fhexpix);
    inc(i);
  end;
  cursorpos.endh.y := fhexarea.top + (fcurend div fcharscount - fstart)* fcharheight;

  cursorpos.text.x := ftextarea.left + fcurstart mod fcharscount * fcharwidth;
  cursorpos.text.y := ftextarea.top + (fcurstart div fcharscount - fstart)* fcharheight;
  cursorpos.endt.x := ftextarea.left + fcurend mod fcharscount * fcharwidth;
  cursorpos.endt.y := ftextarea.top + (fcurend div fcharscount - fstart)* fcharheight;
end;

procedure TXSHexView.UpdateArea;
begin
  if floaded then begin
    fcharwidth  := Canvas.TextWidth('W');
    fcharheight := Canvas.TextHeight('W');
    fgutterwidth:= 5+fcharwidth*8+5;
    fhexpix     := fcharwidth*2+fpix;
    SetFocus;
    floaded     := false;
  end;
  fcharscount := 0;
  fvisibleline:= (height-fcharheight-fborder) div fcharheight;
  fheadarea   := Rect(0,0,width-VScroll.width-1,fcharheight);
  fgutterarea := Rect(0,fcharheight,fgutterwidth,height-1);
  fhexarea    := Rect(fgutterwidth+fborder,fcharheight+fborder,fgutterwidth+fborder,fgutterarea.bottom);
  ftextarea   := Rect(width-VScroll.width-fborder-1,fhexarea.top,width-VScroll.width-fborder-1,fhexarea.bottom);
  while fhexarea.right<ftextarea.left-fcharwidth-fhexpix do begin
    if fcharscount mod 8 = 0 then inc(fhexarea.right,fcharwidth);
    inc(fhexarea.right,fhexpix);
    dec(ftextarea.left,fcharwidth);
    inc(fcharscount);
  end;
end;

procedure TXSHexView.Resize;
begin
  inherited Resize;
  UpdateArea;
  UpdateCursor;
end;

procedure TXSHexView.Paint;
var
  i,x,y,a : integer;
  startsel: integer;
  endsel  : integer;
  flag,sel: boolean;
  found   : boolean;
  old     : integer;
  bmp     : TBitmap;

  function HiByte(b: byte): char;
  const s = '0123456789ABCDEF';
  begin
    result := s[b shr 4 + 1];
  end;

  function LoByte(b: byte): char;
  const s = '0123456789ABCDEF';
  begin
    b := b shl 4;
    result := s[b shr 4 + 1];
  end;

begin

  bmp     := TBitmap.Create;
  bmp.width := width;
  bmp.height := height;
  bmp.Canvas.Font := Canvas.Font;
  bmp.Canvas.Brush.Style := bsSolid;
  bmp.Canvas.Pen.Color:=ClWhite;
  bmp.Canvas.Brush.Color:=ClWhite;
  bmp.Canvas.FillRect(bmp.Canvas.ClipRect);

  bmp.Canvas.Brush.Style := bsSolid;
  bmp.Canvas.Pen.Color   := cl3DLight;
  bmp.Canvas.Brush.Color := cl3DLight;
  bmp.Canvas.Rectangle(fheadarea);
  bmp.Canvas.Rectangle(fgutterarea);

  bmp.Canvas.Pen.Color := clSilver;
  bmp.Canvas.Line(fgutterarea.right,fheadarea.top,fhexarea.right-fborder,fheadarea.top);
  bmp.Canvas.Line(fgutterarea.right,fheadarea.bottom,fheadarea.right,fheadarea.bottom);
  bmp.Canvas.Line(fgutterarea.right,fgutterarea.top,fgutterarea.right,fgutterarea.bottom-1);
  bmp.Canvas.Line(fhexarea.right-fborder,fheadarea.bottom,fhexarea.right-fborder,height-1);

  bmp.Canvas.Brush.Style := bsClear;
  bmp.Canvas.Font.Color  := clGray;
  bmp.Canvas.TextOut(5+fcharwidth,4,'Offset');
  bmp.Canvas.TextOut(ftextarea.left+(ftextarea.right-ftextarea.left-bmp.Canvas.TextWidth('Text')) div 2,0,'Text');
  i := 0;
  for x:=0 to fcharscount-1 do begin
    if (x>0) and (x mod 8 = 0) then inc(i,fcharwidth);
    bmp.Canvas.Font.Color  := clGray;
    bmp.Canvas.TextOut(fhexarea.left+x*fhexpix+i,fheadarea.top,IntToHex(x,2));
  end;
  for y:=0 to fvisibleline-1 do
    bmp.Canvas.TextOut(5,fheadarea.bottom+fborder+y*fcharheight,IntToHex((y+fstart)*fcharscount,8));

  if fcharscount<=0 then begin
    BitBlt(Canvas.handle,0,0,width,height,bmp.canvas.handle,0,0,SRCCOPY);
    bmp.free;
    exit;
  end;

  Swap(startsel,endsel);
  sel := startsel<>endsel;

  found := false;
  y := 0;
  while (y < fvisibleline) and not found do begin
    i := 0;
    x := 0;

    if sel then begin

      bmp.Canvas.Brush.Style := bsSolid;

      if fhexactive then begin
        bmp.Canvas.Pen.Color   := $A56D53;
        bmp.Canvas.Brush.Color := $A56D53;
      end else begin
        bmp.Canvas.Pen.Color   := $CCCCCC;
        bmp.Canvas.Brush.Color := $CCCCCC;
      end;

      if (y+fstart = startsel div fcharscount) and ((startsel div fcharscount) = (endsel div fcharscount)) then begin
        if (fcurstart < fcurend) then begin
          bmp.Canvas.Rectangle(cursorpos.hex.x,cursorpos.hex.y,cursorpos.endh.x+fcharwidth*2,cursorpos.endh.y+fcharheight);
        end else begin
          bmp.Canvas.Rectangle(cursorpos.endh.x,cursorpos.endh.y,cursorpos.hex.x+fcharwidth*2,cursorpos.hex.y+fcharheight);
        end;
      end else if (y+fstart = startsel div fcharscount) then begin
        if (fcurstart < fcurend) then begin
          bmp.Canvas.Rectangle(cursorpos.hex.x,cursorpos.hex.y,fhexarea.right-fcharwidth-fpix,cursorpos.hex.y+fcharheight);
        end else begin
          bmp.Canvas.Rectangle(cursorpos.endh.x,cursorpos.endh.y,fhexarea.right-fcharwidth-fpix,cursorpos.endh.y+fcharheight);
        end;
      end else if (y+fstart = endsel div fcharscount) then begin
        if (fcurstart > fcurend) then begin
          bmp.Canvas.Rectangle(fhexarea.left,cursorpos.hex.y,cursorpos.hex.x+fcharwidth*2,cursorpos.hex.y+fcharheight);
        end else begin
          bmp.Canvas.Rectangle(fhexarea.left,cursorpos.endh.y,cursorpos.endh.x+fcharwidth*2,cursorpos.endh.y+fcharheight);
        end;
      end else if (y+fstart > startsel div fcharscount) and (y+fstart < endsel div fcharscount) then begin
        bmp.Canvas.Rectangle(fhexarea.left,fhexarea.top+y*fcharheight,fhexarea.right-fcharwidth-fpix,fhexarea.top+y*fcharheight+fcharheight);
      end;

      if ftextactive then begin
        bmp.Canvas.Pen.Color   := $A56D53;
        bmp.Canvas.Brush.Color := $A56D53;
      end else begin
        bmp.Canvas.Pen.Color   := $CCCCCC;
        bmp.Canvas.Brush.Color := $CCCCCC;
      end;

      if (y+fstart = startsel div fcharscount) and ((startsel div fcharscount) = (endsel div fcharscount)) then begin
        if (fcurstart < fcurend) then begin
          bmp.Canvas.Rectangle(cursorpos.text.x,cursorpos.text.y,cursorpos.endt.x+fcharwidth,cursorpos.endt.y+fcharheight);
        end else begin
          bmp.Canvas.Rectangle(cursorpos.endt.x,cursorpos.endt.y,cursorpos.text.x+fcharwidth,cursorpos.text.y+fcharheight);
        end;
      end else if (y+fstart = startsel div fcharscount) then begin
        if (fcurstart < fcurend) then begin
          bmp.Canvas.Rectangle(cursorpos.text.x,cursorpos.text.y,ftextarea.right,cursorpos.text.y+fcharheight);
        end else begin
          bmp.Canvas.Rectangle(cursorpos.endt.x,cursorpos.endt.y,ftextarea.right,cursorpos.endt.y+fcharheight);
        end;
      end else if (y+fstart = endsel div fcharscount) then begin
        if (fcurstart > fcurend) then begin
          bmp.Canvas.Rectangle(ftextarea.left,cursorpos.text.y,cursorpos.text.x+fcharwidth,cursorpos.text.y+fcharheight);
        end else begin
          bmp.Canvas.Rectangle(ftextarea.left,cursorpos.endt.y,cursorpos.endt.x+fcharwidth,cursorpos.endt.y+fcharheight);
        end;
      end else if (y+fstart > startsel div fcharscount) and (y+fstart < endsel div fcharscount) then begin
        bmp.Canvas.Rectangle(ftextarea.left,ftextarea.top+y*fcharheight,ftextarea.right,ftextarea.top+y*fcharheight+fcharheight);
      end;

    end;

    while (x < fcharscount) and not found do begin
      a := x+y*fcharscount+fstart*fcharscount;
      if (x>0) and (x mod 8 = 0) then inc(i,fcharwidth);
      if a < flen then begin

        flag := (a>=startsel) and (a<=endsel);


        if (a=fcurstart) then begin

          bmp.Canvas.Brush.Style := bsSolid;
          bmp.Canvas.Pen.Color   := $E6B7B7;
          bmp.Canvas.Brush.Color := $E6B7B7;
          bmp.Canvas.Rectangle(cursorpos.hex.x,cursorpos.hex.y,cursorpos.hex.x+fcharwidth*2,cursorpos.hex.y+fcharheight);
          bmp.Canvas.Pen.Color   := fblinkclr;
          bmp.Canvas.Brush.Color := fblinkclr;
          bmp.Canvas.Rectangle(cursorpos.frame.x,cursorpos.frame.y,cursorpos.frame.x+fcharwidth,cursorpos.frame.y+fcharheight);
          bmp.Canvas.Rectangle(cursorpos.text.x,cursorpos.text.y,cursorpos.text.x+fcharwidth,cursorpos.text.y+fcharheight);

          bmp.Canvas.Brush.Style := bsClear;
          bmp.Canvas.Font.Color := clGreen;
          old := bmp.Canvas.Font.Color;
          if not frframe then bmp.Canvas.Font.Color  := clRed;
          bmp.Canvas.TextOut(fhexarea.left+x*fhexpix+i,fheadarea.bottom+fborder+y*fcharheight,HiByte(fdata[a]));
          bmp.Canvas.Font.Color := old;
          if frframe then bmp.Canvas.Font.Color      := clRed;
          bmp.Canvas.TextOut(fhexarea.left+x*fhexpix+i+fcharwidth,fheadarea.bottom+fborder+y*fcharheight,LoByte(fdata[a]));

          bmp.Canvas.Font.Color  := clRed;
          if not (fdata[a] in [0..31]) then
            bmp.Canvas.TextOut(ftextarea.left+x*fcharwidth,fheadarea.bottom+fborder+y*fcharheight,chr(fdata[a]))
          else
            bmp.Canvas.TextOut(ftextarea.left+x*fcharwidth,fheadarea.bottom+fborder+y*fcharheight,'.');

        end else begin

          bmp.Canvas.Brush.Style := bsClear;

          if fhexactive then
            bmp.Canvas.Font.Color  := clBlack
          else
            bmp.Canvas.Font.Color  := clGray;
          if flag and sel then
            bmp.Canvas.Font.Color  := clWhite;

          if ChangedArray[a] then bmp.Canvas.Font.Color  := clRed;

          bmp.Canvas.TextOut(fhexarea.left+x*fhexpix+i,fhexarea.top+y*fcharheight,IntToHex(fdata[a],2));

          if ftextactive then
            bmp.Canvas.Font.Color  := clBlack
          else
            bmp.Canvas.Font.Color  := clGray;
          if flag and sel then
            bmp.Canvas.Font.Color  := clWhite;

          if ChangedArray[a] then bmp.Canvas.Font.Color  := clRed;

          if not (fdata[a] in [0..31]) then
            bmp.Canvas.TextOut(ftextarea.left+x*fcharwidth,fheadarea.bottom+fborder+y*fcharheight,chr(fdata[a]))
          else
            bmp.Canvas.TextOut(ftextarea.left+x*fcharwidth,fheadarea.bottom+fborder+y*fcharheight,'.');

        end;

      end else
        found := true;
      inc(x);
    end;
    inc(y);
  end;

  VScroll.Max      := flen div fcharscount + 1;
  VScroll.PageSize := fvisibleline;
  //bmp.Canvas.TextOut(0,0,IntToStr(fstart)+' '+IntToStr(flen div fcharscount - fvisibleline+1));
  BitBlt(Canvas.handle,0,0,width,height,bmp.canvas.handle,0,0,SRCCOPY);
  bmp.free;

end;

end.

